from pydantic import BaseModel
from decimal import Decimal, InvalidOperation

from .prompt import JinjaStringTemplate
from .parser import JSONParser
from .chat import Chat


def ai(fn):
    """
    Decorator for creating LLM-powered functions.

    The function docstring will be used as a Jinja2 template, rendered with
    the function arguments as context, and used as a prompt for the LLM.

    The function arguments and return value must be annotated with type hints.
    The return value may be a Pydantic model, in which case the LLM output
    will be parsed as JSON and validated against the model's schema. It may
    also be one of (int, float, bool, dict, Decimal), in which case the LLM
    output will be parsed as that type. If the return value is annotated as
    a string or some other type, no parsing will occur.

    The function body may be empty or contain "alignment code" that will be
    executed after the LLM output is received. The alignment code may raise
    an assertion error, which will be used to prompt the LLM to fix its output
    and try again.

    The function body is only executed if a special `_retval` argument is
    present in the function definition. This argument will be set to the parsed
    LLM output. If the argument is present and the function code is executed,
    it should return a vaule that will be used as the function's return value,
    which may or may not be what the LLM returned.

    If the argument is not present and the function code is not executed,
    the parsed LLM output will be returned automatically.

    :param fn: The function to decorate.
    :return: A decorated function.
    """

    if not fn.__doc__ or not fn.__doc__.strip():
        raise ValueError("The function must have a docstring")

    parser = None
    return_type = fn.__annotations__.get("return")
    return_type_prompt = ""

    # If the return type is something we know how to parse, set up the parser
    # and instructions for the LLM so it knows how it should format its output.
    if return_type:
        if issubclass(return_type, BaseModel):
            parser = JSONParser(spec=return_type)
            return_type_prompt = f"Your output must obey the following JSON schema: {return_type.model_json_schema()}"
        elif return_type in [int, float, bool, Decimal]:

            def parser(output):
                try:
                    return return_type(output)
                except (ValueError, TypeError, InvalidOperation) as err:
                    raise ValueError(
                        f"Error parsing '{output}' as {return_type.__name__}: {err}"
                    )

            return_type_prompt = (
                f"Your output must be parsable as Python type {return_type.__name__}"
            )

    # Combined parser that runs the LLM output through the parser and then
    # passes it to the function, with both the parser and the function
    # being optional. Using the combined parser allows us to use the built-in
    # LLM parsing logic for both the parser and the alignhment code.
    def align_parser(response, **kwargs):
        if parser:
            response = parser(response)

        if "_retval" not in fn.__annotations__:
            return response

        try:
            return fn(_retval=response, **kwargs)
        except AssertionError as e:
            raise ValueError(
                f"There's a problem with your output: {e}; please try again."
            )

    tpl = JinjaStringTemplate()

    def llm_wrapper(llm, **kwargs) -> str:
        prompt = tpl(fn.__doc__, **kwargs)
        if return_type_prompt:
            prompt += (
                "\n\nOutput ONLY the requested output in a single Markdown code block, without commentary or explanation. "
                + return_type_prompt
            )

        return llm(Chat(prompt), parser=lambda output: align_parser(output, **kwargs))

    return llm_wrapper



from logging import getLogger
from os import getenv
import time
from typing import Callable, Optional

from openai import OpenAI, OpenAIError
from openai.types.chat import ChatCompletion

from ..chat import Chat

log = getLogger(__name__)


class ToolError(Exception):
    def __init__(self, message: str):
        self.message = message


class ChatGPT:
    """
    Client for OpenAI ChatGPT API.
    """

    MODELS = [
        "gpt-3.5-turbo",
        "gpt-3.5-turbo-16k",
        "gpt-4",
        "gpt-4-32k",
        "gpt-4-1106-preview",
        "gpt-4-vision-preview",
    ]

    def __init__(
        self,
        api_key: Optional[str] = None,
        model: str = "gpt-4-1106-preview",
        temperature: float = 0.7,
        timeout: int = 120.0,
        max_retries: int = 3,
    ):
        """
        Create a new ChatGPT client instance.

        :param api_key: OpenAI API key (default: use from OPENAI_API_KEY env var).
        :param model: Model to use (default: gpt-3.5-turbo).
        :param temperature: Temperature parameter (default: 0.7).
        :param timeout: Timeout for API requests (default: 120.0).
        :param max_retries: Maximum number of retries for API requests (default: 3).
        """
        if api_key is None:
            api_key = getenv("OPENAI_API_KEY")

        if not api_key:
            raise ValueError("OpenAI API key is not set")

        if model not in self.MODELS:
            raise ValueError(
                f"Unsupported model: {model} (supported: {','.join(self.MODELS)})"
            )

        self.client = OpenAI(api_key=api_key, timeout=timeout, max_retries=max_retries)
        self.api_key = api_key
        self.model = model
        self.temperature = temperature

    def _run_tool(
        self,
        function_call,
        tools,
    ):
        assert tools, "Tools should not be empty/missing at this point"

        function_name = function_call.name

        available_tools = {t.__name__: t for t in tools}
        available_tool_list = ",".join(available_tools.keys())
        if function_name not in available_tools:
            log.warning(
                f"GPT requested unknown tool: {function_name}; available tools: {available_tool_list}"
            )
            return f"ERROR: Unknown tool: {function_name}; available tools: {available_tool_list}"

        t = available_tools[function_name]
        try:
            args = t._validate_arguments(function_call.arguments)
        except TypeError as err:
            log.warning(
                f"Invalid arguments for GPT tool {function_name}: {err}", exc_info=True
            )
            return f"ERROR: {err}"

        try:
            log.debug(f"Running GPT tool {function_name} with args: {args}")
            return t(**args)
        except Exception as err:
            log.warning(f"Error running GPT tool {function_name}: {err}", exc_info=True)
            raise ToolError(f"Error running GPT tool {function_name}: {err}") from err

    def _call_chatgpt(
        self,
        messages: list[dict[str, str]],
        tools: Optional[list] = None,
    ) -> ChatCompletion:
        try:
            log.debug(f"Calling ChatGPT with messages: {messages})")
            api_kwargs = dict(
                model=self.model,
                temperature=self.temperature,
                messages=messages,
            )
            if tools:
                functions = [t._get_json_schema() for t in tools]
                api_kwargs["functions"] = functions
                api_kwargs["function_call"] = "auto"

            t0 = time.time()
            api_result = self.client.chat.completions.create(**api_kwargs)
            t1 = time.time()
            log.debug(
                f"ChatGPT request completed in {t1 - t0:.2f}s, {api_result.usage.total_tokens} tokens used"
            )
            return api_result.choices[0].message
        except OpenAIError as err:
            log.warning(f"Error calling ChatGPT: {err}", exc_info=True)
            raise

    def __call__(
        self,
        chat: Chat,
        tools: Optional[list] = None,
        parser: Optional[Callable] = None,
        max_iterations: int = 5,
    ) -> Optional[str]:
        chat = chat.fork()

        for i in range(max_iterations):
            response = self._call_chatgpt(list(chat), tools)
            if response.function_call:
                chat.assistant(f"Using tool '{response.function_call.name}'")
                try:
                    result = self._run_tool(response.function_call, tools)
                except ToolError as err:
                    print(repr(err.__cause__), type(err.__cause__))
                    raise err.__cause__
                chat.function(result, name=response.function_call.name)
                continue

            if parser:
                try:
                    content = parser(response.content)
                except ValueError as err:
                    log.debug(f"Error parsing GPT response: {err}", exc_info=True)
                    chat.assistant(response.content)
                    chat.user(
                        f"Error parsing response: {err}. Please output your response EXACTLY as requested."
                    )
                    continue
            else:
                content = response.content

            return content

        return None



from enum import Enum
import json
import re
from typing import Union, Optional

from pydantic import BaseModel


class MultiCodeBlockParser:
    """
    Parse multiple Markdown code blocks from a string.

    Expects zero or more blocks, and ignores any text
    outside of the code blocks.

    Example usage:

    >>> parser = MultiCodeBlockParser()
    >>> text = '''
    ... text outside block
    ...
    ... ```python
    ... first block
    ... ```
    ... some text between blocks
    ... ```js
    ... more
    ... code
    ... ```
    ... some text after blocks
    '''
    >>> assert parser(text) == ["first block", "more\ncode"]

    If no code blocks are found, an empty list is returned:
    """

    def __init__(self):
        self.pattern = re.compile(r"```([a-z0-9]+\n)?(.*?)```\s*", re.DOTALL)

    def __call__(self, text: str) -> list[str]:
        blocks = []
        for block in self.pattern.findall(text):
            blocks.append(block[1].strip())
        return blocks


class CodeBlockParser(MultiCodeBlockParser):
    """
    Parse a Markdown code block from a string.

    Expects exactly one code block, and ignores
    any text before or after it.

    Usage:
    >>> parser = CodeBlockParser()
    >>> text = "text\n```py\ncodeblock\n'''\nmore text"
    >>> assert parser(text) == "codeblock"

    This is a special case of MultiCodeBlockParser,
    checking that there's exactly one block.
    """

    def __call__(self, text: str) -> str:
        blocks = super().__call__(text)
        if len(blocks) != 1:
            raise ValueError(f"Expected a single code block, got {len(blocks)}")
        return blocks[0]


class JSONParser:
    def __init__(self, spec: Optional[BaseModel] = None, strict: bool = True):
        self.spec = spec
        self.strict = strict or (spec is not None)

    @property
    def schema(self):
        return self.spec.model_json_schema() if self.spec else None

    def __call__(self, text: str) -> Union[BaseModel, dict, None]:
        text = text.strip()
        if text.startswith("```"):
            try:
                text = CodeBlockParser()(text)
            except ValueError:
                if self.strict:
                    raise
                else:
                    return None

        try:
            data = json.loads(text.strip())
        except json.JSONDecodeError as e:
            if self.strict:
                raise ValueError("Could not parse JSON") from e
            else:
                return None

        if self.spec is None:
            return data

        try:
            model = self.spec(**data)
        except Exception as err:
            raise ValueError(f"Error parsing JSON: {err}")

        return model


class EnumParser:
    def __init__(self, spec: Enum, ignore_case: bool = True):
        self.spec = spec
        self.ignore_case = ignore_case

    def __call__(self, text: str) -> Enum:
        text = text.strip()
        if self.ignore_case:
            text = text.lower()
        try:
            return self.spec(text)
        except ValueError as e:
            options = ", ".join([str(v) for v in self.spec])
            raise ValueError(
                f"Invalid option '{text}'; valid options: {options}"
            ) from e



from pydantic import Field, create_model, ValidationError
from typing import Callable, Union
import re
import inspect
import json


def parse_docstring(docstring: str) -> tuple[str, dict[str, str]]:
    """
    Parse function arguments from reST docstring

    :param docstring: The text to parse in reST format
    :returns: The description and the dict of arguments and their descriptions
    """

    LINE_MATCH = re.compile(r"^:((param (\w+))|(returns)): (.+)$")

    desc = []
    args = {}
    returns = None

    if docstring is None:
        return "", {}

    for line in docstring.split("\n"):
        if not line:
            continue

        line = line.strip()
        m = LINE_MATCH.match(line)
        if m:
            arg_name = m.group(3)
            description = m.group(5).strip()
            if arg_name:
                args[arg_name] = description
            else:
                returns = description
        elif line:
            desc.append(line)

    desc = " ".join(desc)
    if returns:
        if desc:
            desc += "\n"
        desc += f"Returns {returns}"
    return desc, args


def create_pydantic_model_from_function(fn: Callable) -> type:
    """
    Inspect a function and create a Pydantic model from its arguments,
    return type and docstring.

    The model will be named after the function, have the docstring
    as its description and the function arguments as its fields.
    The function arguments must have type annotations and the docstring
    must contain reST formatted descriptions of the arguments.

    :param fn: The function to inspect
    :returns: The newly created Pydantic model
    """
    fields = {}

    desc, param_docs = parse_docstring(fn.__doc__)

    signature = inspect.signature(fn)
    for param_name, param in signature.parameters.items():
        if param_name == "self":
            continue

        param_doc = param_docs.get(param_name, "")
        param_default = param.default

        if param_default is not inspect._empty:
            field_def = Field(param_default, description=param_doc)
        else:
            field_def = Field(..., description=param_doc)

        if param.annotation is inspect._empty:
            raise TypeError(
                f"Parameter {param_name} of {fn.__name__} has no type annotation"
            )
        fields[param_name] = (param.annotation, field_def)

    model = create_model(
        fn.__name__,
        **fields,
        __cls_kwargs__={"extra": "forbid"},
    )
    model.__doc__ = desc
    return model


def tool(fn: Union[Callable, str]) -> Callable:
    """
    Decorator to mark a function as a tool

    Usage:

    >>> @tool
    >>> def my_tool(a: int, b: int = 2) -> int:
    >>>     '''
    >>>     My tool description
    >>>
    >>>     :param a: First argument
    >>>     :param b: Second argument
    >>>     :returns: Something
    >>>     '''
    >>>     return a + b

    The tool can also be named explicitly:

    >>> @tool("add_two_numbers")
    >>> def my_tool(a: int, b: int = 2) -> int:
    >>>     '''
    >>>     :param a: First argument
    >>>     :param b: Second argument
    >>>     :returns: Something
    >>>     '''
    >>>     return a + b
    """

    def decorator(fn: Callable, tool_name: str) -> Callable:
        model = create_pydantic_model_from_function(fn)

        def validate(json_args):
            try:
                kwargs = json.loads(json_args)
            except json.JSONDecodeError as err:
                raise TypeError(f"Arguments are not valid JSON: {err}")

            try:
                return model(**kwargs).model_dump()
            except ValidationError as err:
                raise TypeError(f"Invalid arguments: {err.errors()}")

        def get_schema():
            schema = model.model_json_schema()
            name = fn.__name__
            desc = schema.pop("description", "")
            return {
                "name": name,
                "description": desc,
                "parameters": schema,
            }

        fn._validate_arguments = validate
        fn._get_json_schema = get_schema
        fn.__name__ = tool_name

        schema = get_schema()
        fn._oneline_description = f"{schema['name']} - {schema['description']}"
        return fn

    if callable(fn):
        return decorator(fn, fn.__name__)
    else:
        return lambda func: decorator(func, fn)



from copy import deepcopy
from typing import Iterator, Optional


class Chat:
    """
    A conversation between a user and a Large Language Model (LLM) assistant.
    """

    ROLES = ["system", "user", "assistant", "function"]

    messages: list[dict[str, str]]

    def __init__(self, content: Optional[str] = None):
        """
        Initialize a new conversation.

        :param content: Initial system message (optional).
        """
        self.messages = []
        if content is not None:
            self.system(content)

    @staticmethod
    def _dedent(text: str) -> str:
        """
        Remove common leading whitespace from every line of text.

        :param text: Text to dedent.
        :returns: Dedented text.
        """
        indent = len(text)
        lines = text.splitlines()
        for line in lines:
            if line.strip():
                indent = min(indent, len(line) - len(line.lstrip()))
        dedented_lines = [line[indent:].rstrip() for line in lines]
        return "\n".join(line for line in dedented_lines)

    def add(self, role: str, content: str, name: Optional[str] = None) -> "Chat":
        """
        Add a message to the conversation.

        In most cases, you should use the convenience methods instead.

        :param role: Role of the message (system, user, assistant, function).
        :param content: Content of the message.
        :param name: Name of the message sender (optional).
        :returns: The chat object.
        """

        if role not in self.ROLES:
            raise ValueError(f"Unknown role: {role}")
        if not content:
            raise ValueError("Empty message content")
        if not isinstance(content, str) and not isinstance(content, dict):
            raise TypeError(f"Invalid message content: {type(content).__name__}")

        message = {
            "role": role,
            "content": self._dedent(content) if isinstance(content, str) else content,
        }
        if name is not None:
            message["name"] = name

        self.messages.append(message)
        return self

    def system(self, content: str, name: Optional[str] = None) -> "Chat":
        """
        Add a system message to the conversation.

        System messages can use `name` for showing example conversations
        between an example user and an example assistant.

        :param content: Content of the message.
        :param name: Name of the message sender (optional).
        :returns: The chat object.
        """
        return self.add("system", content, name)

    def user(self, content: str, name: Optional[str] = None) -> "Chat":
        """
        Add a user message to the conversation.

        :param content: Content of the message.
        :param name: User name (optional).
        :returns: The chat object.
        """
        return self.add("user", content, name)

    def assistant(self, content: str, name: Optional[str] = None) -> "Chat":
        """
        Add an assistant message to the conversation.

        :param content: Content of the message.
        :param name: Assistant name (optional).
        :returns: The chat object.
        """
        return self.add("assistant", content, name)

    def function(self, content: str, name: Optional[str] = None) -> "Chat":
        """
        Add a function (tool) response to the conversation.

        :param content: Content of the message.
        :param name: Function/tool name (optional).
        :returns: The chat object.
        """
        return self.add("function", content, name)

    def fork(self) -> "Chat":
        """
        Create an identical copy of the conversation.

        This performs a deep copy of all the message
        contents, so you can safely modify both the
        parent and the child conversation.

        :returns: A copy of the conversation.
        """
        child = Chat()
        child.messages = deepcopy(self.messages)
        return child

    def after(self, parent: "Chat") -> "Chat":
        """
        Create a chat with only messages after the last common
        message (that appears in both parent conversation and
        this one).

        :param parent: Parent conversation.
        :returns: A new conversation with only new messages.
        """
        index = 0
        while (
            index < min(len(self.messages), len(parent.messages))
            and self.messages[index] == parent.messages[index]
        ):
            index += 1

        child = Chat()
        child.messages = [deepcopy(msg) for msg in self.messages[index:]]
        return child

    def last(self) -> Optional[dict[str, str]]:
        """
        Get the last message in the conversation.

        :returns: The last message, or None if the conversation is empty.
        """
        return self.messages[-1] if self.messages else None

    def __iter__(self) -> Iterator[dict[str, str]]:
        """
        Iterate over the messages in the conversation.

        :returns: An iterator over the messages.
        """
        return iter(self.messages)

    def __repr__(self) -> str:
        return f"<Chat({self.messages})>"



from copy import deepcopy
from logging import getLogger
from os import getenv
from typing import Optional, Callable

from anthropic import Anthropic, HUMAN_PROMPT, AI_PROMPT, BadRequestError

from ..chat import Chat

log = getLogger(__name__)


class Claude:
    MODELS = [
        "claude-2",
        "claude-instant-1",
    ]

    def __init__(
        self,
        api_key: str,
        model: str = "claude-2",
    ):
        if api_key is None:
            api_key = getenv("ANTHROPIC_API_KEY")

        if not api_key:
            raise ValueError("Antrophic API key must be provided")

        if model not in self.MODELS:
            raise ValueError(f"Unsupported model: {model}")

        self.api_key = api_key
        self.model = model
        self.client = Anthropic(api_key=self.api_key)

    @staticmethod
    def _sanitize_message(txt: str) -> str:
        """
        Sanitize content so there's no prompt in the middle of the message.

        After each iteration, the function rechecks whether the prompt is still
        in the message and loops if it is. This is necessary to handle cases like:

        "\n\nHu\n\nHuman:man"

        ...which would be sanitized to: "\n\nHuman" the first time, allowing prompt
        injection.
        """
        while True:
            new_txt = txt.replace(HUMAN_PROMPT, "").replace(AI_PROMPT, "")
            if new_txt == txt:
                return new_txt
            txt = new_txt

    @classmethod
    def get_text_prompt(cls, chat: Chat) -> str:
        prompt = HUMAN_PROMPT + " "
        speaker_is_human = True
        for message in chat:
            if message["role"] in ("system", "user"):
                if not speaker_is_human:
                    prompt += HUMAN_PROMPT + " "
                    speaker_is_human = True
            elif message["role"] == "assistant":
                if speaker_is_human:
                    prompt += AI_PROMPT + " "
                    speaker_is_human = False
            else:
                raise ValueError(f"Unsupported message role: {message['role']}")

            prompt += cls._sanitize_message(message["content"]) + "\n"

        # Prime the assistant to reply
        prompt += AI_PROMPT

        # Our newlines + human/assistant prompt newlines are too many, remove the
        # extra ones just before the speaker change.
        prompt = prompt.replace("\n\n\n", "\n\n").rstrip()

        return prompt

    @staticmethod
    def inject_tools(chat: Chat, tools: Optional[list]) -> Chat:
        """
        Inject tool messages into the chat.

        This is necessary because Claude doesn't support tool injection yet.
        """
        if not tools:
            return chat

        new_chat = chat.fork()
        new_chat.messages = new_chat.messages[:1]

        descriptions = "\n".join([f"* {t._oneline_description}" for t in tools])
        new_chat.system(
            """
You have access to the following tools:
%s

To use a tool, return a message in the following JSON format:
{
    "action": "USE_TOOL",
    "name": "<tool-name>",
    "arguments": {
        "arg1": "value1",
        "arg2": "value2",
        ...
    }
}
"""
            % descriptions
        )
        new_chat.messages.extend(deepcopy(chat.messages[1:]))
        return new_chat

    def __call__(
        self,
        chat: Chat,
        max_tokens: int = 1000,
        tools: Optional[list] = None,
        parser: Optional[Callable] = None,
        max_iterations: int = 5,
    ) -> Optional[str]:
        if parser:
            raise NotImplementedError("Custom parsers not implemented yet")

        if tools:
            chat = self.inject_tools(chat, tools)

        prompt = self.get_text_prompt(chat)
        log.debug(f"Calling Claude with prompt: {prompt}")

        try:
            response = self.client.completions.create(
                model=self.model,
                prompt=prompt,
                max_tokens_to_sample=max_tokens,
            )
        except BadRequestError as err:
            log.warning(f"Error calling Claude: {err}", exc_info=True)
            return None

        return response.completion



from pathlib import Path
from typing import Any, Optional

from jinja2 import Environment, BaseLoader, FileSystemLoader, StrictUndefined


class FormatTemplate:
    def __call__(self, template: str, **kwargs: dict[str, Any]) -> str:
        return template.format(**kwargs)


class BaseJinjaTemplate:
    def __init__(self, loader: Optional[BaseLoader]):
        self.env = Environment(
            loader=loader,
            autoescape=False,
            lstrip_blocks=True,
            trim_blocks=True,
            keep_trailing_newline=True,
            undefined=StrictUndefined,
        )


class JinjaStringTemplate(BaseJinjaTemplate):
    def __init__(self):
        super().__init__(None)

    def __call__(self, template: str, **kwargs: dict[str, Any]) -> str:
        tpl = self.env.from_string(template)
        return tpl.render(**kwargs)


class JinjaFileTemplate(BaseJinjaTemplate):
    def __init__(self, template_dir: str):
        if not Path(template_dir).is_dir():
            raise ValueError(f"Template directory does not exist: {template_dir}")
        super().__init__(FileSystemLoader(template_dir))

    def __call__(self, template: str, **kwargs: dict[str, Any]) -> str:
        tpl = self.env.get_template(template)
        return tpl.render(**kwargs)
